# -*- coding: utf-8 -*-
"""
Specifies modular pre-written motion and dispense sequences for common prints.

| First created on Sun Oct 20 00:08:15 2019
| Revised: 20/10/2019 00:34:27
| Author: Bijal Patel

"""
import sys
from abc import ABC, abstractmethod
from polychemprint3.utility.loggerSpec import loggerSpec


class sequenceSpec(loggerSpec, ABC):
    """Abstract Base Class for predefined print sequences."""

    ################### Construct/Destruct METHODS ###########################
    @abstractmethod
    def __init__(self, nameString, descrip, axes, tool, verbose, **kwargs):
        """*Initializes sequence object*.

        Parameters
        ----------
        axes: PCP_Axes object
            Axes object to send motion commands to
        tool: PCP_Tool
            Tool object to send dispense commands to
        verbose: bool
            level of detail to be printed to cmd line
        """
        self.nameString = nameString
        self.descrip = descrip
        self.axes = axes
        self.tool = tool
        self.verbose = verbose
        super().__init__(**kwargs)

    ################### Parameter Methods ###########################
    def stringParams(self):
        """*Turns paramList into a formatted string*.

        Returns
        -------
        String
            Formatted String containing table of param and current values
        """
        outString = ""
        for param in self.paramList:
            outString += "\t%-40s|  %-25s\n" % (self.name, self.value)

    ################### Sequence Actions ###################################
    @abstractmethod
    def operateSeq(self):
        """*Performs print sequence*.

        Returns
        -------
        bool
            Whether sequence successfully completed or not
        """
        pass

    @abstractmethod
    def genSequence(self):
        """*Loads print sequence into a list into cmdList attribute*.

        Returns
        -------
        bool
            whether successfully reached the end or not
        """
        pass

    ####################### Logging METHODS ###############################
    @abstractmethod
    def writeLogSelf(self):
        """*Generates log string containing dict to be written to log file*.

        Returns
        -------
        String
            log in string format
        """
        return super().writeLogSelf()

    @abstractmethod
    def loadLogSelf(self, logString):
        """*loads log back into dict*.

        Parameters
        ----------
        logString: String
            log string to be loaded back in

        """
        super().loadLogSelf(logString)


class seqParam():
    """Base Class for parameters used in sequences."""

    ################### Construct/Destruct METHODS ###########################
    def __init__(self, name, value, unit, helpString):
        """*Initializes parameter object*.

        Parameters
        ----------
        name: String
            name of parameter (e.g., length)
        value
            value of parameter (e.g., 20)
        unit
            unit value is measured in (e.g., mm)
        helpString: String
            Description of any limits/ bounds on the value (e.g., <30)
        """
        self.name = name
        self.value = value
        self.unit = unit
        self.helpString = helpString



# -*- coding: utf-8 -*-
"""Contains toolSpec Abstract Base Class.

| First created on Sun Oct 20 00:03:21 2019
| Revised: 13/11/2019 14:29:53
| Author: Bijal Patel

"""

from abc import ABC, abstractmethod
from polychemprint3.utility.loggerSpec import loggerSpec


class toolSpec(loggerSpec, ABC):
    """Interface for all dispensing/writing tools."""

################### Construct/Destruct METHODS ###########################
    def __init__(self, name, units, **kwargs):
        """*Initializes Tool Object*.

        Parameters
        ----------
        name: String
            tool name
        units: String
            Units this tool operates in

        """
        self.name = name
        self.units = units
        super().__init__(**kwargs)

############################# Activate ### ###############################
    @abstractmethod
    def engage(self):
        """*Activate tool (dispense/LASER on, etc)*
        | *Parameters*
        |   none

        | *Returns*
        |   [1, "Tool Engaged"]
        |   [-1, "Error: Tool could not be engaged"]
        """
        pass

    @abstractmethod
    def disengage(self) :
        """*Deactivates tool (stops dispense/LASER off, etc)*
        | *Parameters*
        |   none

        | *Returns*
        |   [1, "Tool Disengaged"]
        |   [-1, "Error: Tool could not be disengaged"]
        """
        pass

    @abstractmethod
    def setValue(self, value):
        """*Set Tool value of a specified Tool parameter*
        | *Parameters*
        |   value - the new value of the parameter

        | *Returns*
        |   [1, "Value Set succesfully"]
        |   [-1, "Error: Parameter could not be set for Tool + error text"]
        """
        pass

    @abstractmethod
    def getState(self):
        """*Returns active state of tool*
        | *Parameters*
        |   none

        | *Returns*
        |   [1, "Tool On"]
        |   [0, "Tool Off"]
        |   [-1, "Error: Tool activation state cannot be determined + Error]
        """
        pass

####################### Logging METHODS ###############################
    @abstractmethod
    def writeLogSelf(self):
        """*Generates json string containing dict to be written to log file*.

        Returns
        -------
        String
            log in json string format
        """
        return super().writeLogSelf()

    @abstractmethod
    def loadLogSelf(self, jsonString):
        """*loads json log back into dict*.

        Parameters
        ----------
        jsonString: String
            json string to be loaded back in

        """
        super().loadLogSelf(jsonString)



# -*- coding: utf-8 -*-
"""Contains ioMenu Abstract Base Class.

| First created on Sun Oct 20 00:03:21 2019
| Revised: 20/10/2019 00:34:27
| Author: Bijal Patel

"""
from abc import ABC, abstractmethod
from colorama import Fore, Style
import sys
from polychemprint3.commandLineInterface.ioElementSpec \
    import ioElementSpec


class ioMenuSpec(ioElementSpec, ABC):
    """Specifies the interface for CLI menus."""

    #####################################################################
    ### Construct/Destruct METHODS
    #####################################################################
    def __init__(self, menuTitle, menuItems, lastCmd="", memCmd="", **kwargs):
        """*Initializes Tool Object*.

        | *Parameters*
        |   menuTitle, name of the menu
        |   menuItems, dictionary of menu options and text

        | *Returns*
        |   none
        """
        self.menuTitle = menuTitle
        self.menuItems = menuItems
        self.lastCmd = lastCmd
        self.memCmd = memCmd
        super().__init__(**kwargs)

    #####################################################################
    ### UI_CLI_IOELement METHODS
    #####################################################################

    def io_Operate(self):
        """*Do the primary purpose of the CLI element*.

        | *Parameters*
        |   none

        | *Returns*
        |   none or flag, optional string describing operation termination
        """
        self.ioMenu_Operate()

    #####################################################################
    ### Unique METHODS
    #####################################################################

    # Operation Methods
    def ioMenu_printMenu(self):
        """*Prints formatted menu options from menuItems dict*.

        | *Parameters*
        |   none

        | *Returns*
        |   none
        """
        print("-" * 120)
        print("###\t" + self.menuTitle)
        print("-" * 120)
        for key in sorted(self.menuItems):
            print("\t%-40s|  %-25s" % (key, self.menuItems.get(key)))

        print(Style.RESET_ALL)

        storedCmds = {Fore.LIGHTCYAN_EX
                      + "[/] Repeat Last Command": self.lastCmd,
                      Fore.LIGHTCYAN_EX
                      + "[.] Repeat Saved Command": self.memCmd,
                      Fore.LIGHTCYAN_EX
                      + "[,] Store Saved Command": "Will Prompt for command"}

        for key in storedCmds:
            print("\t%-40s|  %-25s" % (key, storedCmds.get(key)))

        print(Style.RESET_ALL)

    def ioMenu_updateStoredCmds(self, lastCmd, memCmd):
        """*Updates stored commands local to this menu item from inputs*.

        | *Parameters*
        |   lastCmd, String specifying the last command entered
        |   memCmd, String specifying the command saved to memory

        | *Returns*
        |   none
        """
        self.lastCmd = lastCmd
        self.memCmd = memCmd

    @abstractmethod
    def ioMenu_Operate(self):
        """*Perform menu operations*.

        | *Parameters*
        |   none

        | *Returns*
        |   flag, String with title of next menu to call
        """



# -*- coding: utf-8 -*-
"""
The *axes3D* class contains all data and methods related to controlling 3D printer axes

| First created on Sat Oct 19 20:39:58 2019
| Revised: 23/10/2019 14:06:59
| Author: Bijal Patel

"""

import serial
import io
import json
from time import time
from polychemprint3.utility.serialDeviceSpec import serialDeviceSpec
from polychemprint3.utility.loggerSpec import loggerSpec

class axes3D(serialDeviceSpec, loggerSpec):
    def __init__(self, devAddress="unset",firmwareVers="unset", baudRate="unset", commsTimeOut=0.5, verbose =0):
        """*Initializes Axes3D object*.

        | *Parameters* All default to "unset"
        |   devAddress - device address on this computer
        |   firmwareVers - version of the firmware loaded on the printer
        |   baudRate - baud rate
        |   commsTimeOut - how long to wait for serial device
        |   verbose - whether details should be printed to cmd line

        | *Returns*
        |   none
        """
        self.devAddress = devAddress
        self.firmwareVers = firmwareVers
        self.baudRate = baudRate
        self.verbose = verbose
        self.commsTimeOut = commsTimeOut

    def move(self, gcodeString):
        """*Initializes Axes3D object*.

        Parameters
        ----------
        gCodeString: String
            Motion command in terms of Gcode G0/G1/G2/G3 supported

        | *Returns*
        |   none
        """
        self.writeReady(gcodeString)



    def checkIfSerialConnectParamsSet(self):
        """*Goes through connection parameters and sees if all are set*.

        | *Parameters*
        |   none

        | *Returns*
        |   True if all parameters are set, false if any unset
        """
        connectParam = [self.devAddress,self.firmwareVers,self.baudRate]
        return 'unset' not in connectParam



    def startSerial(self):
        """*Attempts to connect to Hardware and send an initial handshake*

        | *Parameters*
        |   none

        | *Returns*
        |   [1, [list of read lines] if succesfully able to connect and handshake
        |   [0, 'Not all connection param set'] if not all connection param set
        |   [-1, 'Failed Creating pySerial'] if failed creating pyserial device
        |   [-2, 'Failed Handshake'] if handshake failed
        """

        if self.checkIfSerialConnectParamsSet():
            #Try to connect, catch errors and return to user
            try:
                self.ser = serial.Serial(port = self.devAddress,\
                                         baudrate = self.baudRate,\
                                         bytesize = serial.EIGHTBITS,\
                                         parity = serial.PARITY_NONE,\
                                         stopbits = serial.STOPBITS_ONE,\
                                         timeout = 1,\
                                         xonxoff = False,\
                                         rtscts = False,\
                                         dsrdtr = False,\
                                         writeTimeout = 2
                                         )
                #Use ser for writing
                #Use sReader for buffered read

                self.sReader = io.TextIOWrapper(io.BufferedReader(self.ser))

                #Clear initial garbage text in output buffer
                self.ser.reset_output_buffer()

                time.sleep(0.25)
                lineIn = self.sReader.readlines();
                linesIn = [lineIn]

                #keep reading until empty
                while lineIn != []:
                    time.sleep(0.25)
                    lineIn = self.sReader.readlines();
                    linesIn.extend(lineIn)

            except Exception as inst:
                return [-1, 'Failed Creating pySerial... ' + inst.__str__()]

        else: #Not all params were set
            return [0, 'Not all connection parameters set']

        #Try initial handshake
        handShakeResponse = self.handShakeSerial()
        if (handShakeResponse[0]==1):
            return [1,linesIn]
        else:
            return [-2,handShakeResponse[1]]

    def handShakeSerial(self):
        """*Attempts to 'handshake' with printer* - see if correct firmware version returned

        | *Parameters*
        |   none

        | *Returns*
        |   [1, 'Handshake Success'] if succesfull 2-way communication
        |   [0, 'Handshake Failed, Received: + message received'] if unexpected input received
        |   [-1, 'Handshake failed, Error: + errortext] if exception raised or not correct firmware vers
        """
        try:
            self.ser.write(b"M115\n")
            readInput = self.sReader.readlines()
            print(readInput);
            wrongVers = (-1 == readInput[0].find(self.firmwareVers))
            if (wrongVers):
                return [-1,'Wrong Firmware Version DO NOT CONTINUE']
            else:
                return [1, 'Handshake Success']
        except Exception as inst:
            return [-1, 'Error on Handshake: ' + inst.__str__()]


    def __writeSerial__(self,command):
        """*Writes command to serial device*

        | *Parameters*
        |   command, the string to send

        | *Returns*
        |   [1, 'Command Sent + command'] if succesfull 2-way communication
        |   [0, 'Write Failed + Error'] if exception caught
        """
        try:
            self.ser.write(command)
            if (self.verbose):
                print('\tCommand Sent: ' + command)
            return [1,'Command Sent' + command]
        except Exception as inst:
            return [-1, 'Error on Write: ' + inst.__str__()]

    def readTime(self):
        """*Reads in from serial device until timeout*

        | *Parameters*
        |   none

        | *Returns*
        |   inp String of all text read in, empty string if nothing
        """
        inp = '' #input string
        ins = '' #read in
        tEnd = time() + self.commsTimeOut

        #Reads input until timeout
        while (time() < tEnd):
            ins = self.ser.read();
            if (ins != ""):
                inp += ins;

        inp = inp.strip #removes any newlines
        if self.verbose:
            print('\tReceived    : ' + inp + '\n');
        return (inp);


    def waitReady(self):
        """*Checks if taz is ready to receive new command by Looking for "ok" in input*

        | *Parameters*
        |   none

        | *Returns*
        |   inp, String read in while waiting
        """
        notReady = True
        i = 0 #loop increments
        while (notReady):
            inp = self.readTime() #read buffer
            if (i%10==0 and self.verbose):
                print('\tWaiting for axes... ... ...\n')
            if ('ok' in inp):
                notReady = False
        return inp

    def writeReady(self,command):
        """*Writes when ready*

        | *Parameters*
        |   command, string to write to axes

        | *Returns*
        |   none
        """
        self.write(command)
        self.waitReady()


    def getAbsPosXY(self):
        """*Gets the current position (absolute) from taz and returns XYZ coordinates*

        | *Parameters*
        |   command, string to write to axes

        | *Returns*
        |   [X, Y] X and Y positions as strings
        """

        self.writeReady('M114\n')
        m114Call = self.waitReady()
        m114Split = m114Call.split(' ')
        x = m114Split[0][2:]
        y = m114Split[1][2:]
        return[x,y]

    def goToABSXY(self,xAbs,yAbs):
        """*Moves carriage to provided absolute X,Y coordinates*

        returns to relative positioning at end

        | *Parameters*
        |   xAbs, absolute x position to go to
        |   yAbs, absolute y position to go to

        | *Returns*
        |    none
        """
        self.writeReady('G90\n') #set abs positioning
        #Move rapidly to target position
        self.writeReady('G1 F2000 X' + xAbs + ' Y' + yAbs + '\n')
        self.writeReady('G91\n') #return to relative positioning

    def stopSerial(self):
        """*Closes serial devices*

        | *Parameters*
        |   none

        | *Returns*
        |    none
        """
        self.ser.close()
        self.sReader.close()

    ### Logging METHODS
    def writeLogSelf(self):
        """*Generates log string containing dict to be written to log file*.

        Returns
        -------
        String
            log in string format
        """
        return super().writeLogSelf()

    def loadLogSelf(self, logString):
        """*loads log back into dict*.

        Parameters
        ----------
        logString: String
            log string to be loaded back in

        """
        super().loadLogSelf(logString)



# -*- coding: utf-8 -*-
"""
The *userSpec* Class holds data and operations relating to software users such as home directory and software preferences.

| First created on Sat Oct 19 20:39:58 2019
| Revised: 23/10/2019 14:06:59
| Author: Bijal Patel

"""
class userSpec:
    def __init__(self, name="unset", creationDate="unset", homeDirectory="unset", programSettings="unset"):
        """*Initializes User object*

        | *Parameters* All default to "unset"
        |   name - user name (string)
        |   creationDate - date this user was first created
        |   homeDirectory - location of users data folder
        |   programSettings - array of UI options

        | *Returns*
        |   none
        """

        self.name = name
        self.creationDate = creationDate
        self.homeDirectory = homeDirectory
        self.programSettings = programSettings







"""
Implements the Tool base class for a null Tool [no action, returns true].

| First created on 13/11/2019 13:33:28
| Revised: 13/11/2019 13:33:28
| Author: Bijal Patel

"""

##############################################################################
##################### Imports
##############################################################################

from polychemprint3.tools.toolSpec import toolSpec
from polychemprint3.utility.serialDeviceSpec import serialDeviceSpec


class nullTool(toolSpec, serialDeviceSpec):
    """Implements the Tool base class for a null Tool (no action)."""

    ###########################################################################
    ### Construct/Destruct METHODS
    ###########################################################################
    def __init__(self, name="unset", units="null", devAddress="unset",
                 baudRate="unset", commsTimeOut=0.5, verbose=0, **kwargs):
        """*Initializes T_UltimusExtruder Object*.

        Parameters
        ----------
        name: String
            tool name
        devAddress: Strong
            device address on this computer
        baudRate: int
            baud rate
        commsTimeOut: int
            how long to wait for serial device before timeout on reads
        verbose: bool
            whether details should be printed to cmd line
        """
        self.dispenseStatus = 0  # off
        inputs = {"name": name, "units": units,
                  "devAddress": devAddress, "baudRate": baudRate,
                  "commsTimeOut": commsTimeOut, "verbose": verbose}
        kwargs.update(inputs)
        super().__init__(**kwargs)

    ##########################################################################
    ### PCP_TOOL METHODS
    ##########################################################################

    ############################# Activate METHODS ###########################
    def engage(self):
        """*Toggles Dispense on*.

        Returns
        -------
        [1, "Dispense On"]
        [0, "Error: Dispense Already On"]
        [-1, 'Failed engaging dispense ' + inst.__str__()]
        """
        try:
            if self.dispenseStatus == 0:
                self.dispenseStatus = 1
                return [1, "Dispense On"]

            else:
                return [0, "Error: Dispense Already On"]
        except Exception as inst:
            return [-1, 'Failed engaging dispense ' + inst.__str__()]

    def disengage(self):
        """*Toggles Dispense off*.

        Returns
        -------
        [1, "Dispense Off"]
        [0, "Error: Dispense already off"]
        [-1, 'Failed engaging dispense ' + inst.__str__()]
        """
        try:
            if self.dispenseStatus == 1:
                self.dispenseStatus = 0
                return [1, "Dispense Off"]

            else:
                return [0, "Error: Dispense already off"]
        except Exception as inst:
            return [-1, 'Failed disengaging dispense ' + inst.__str__()]

    ############################# Value METHODS ###########################

    def setValue(self, newVal):
        """*Set Value*.

        Parameters
        ----------
        pressureVal: String
            New value to set

        Returns
        -------
        [output of writeSerialCommand]
        [-1, "Error: Pressure could not be set for Extruder + error text"]
        """
        try:
            return [1, 'null mode: newVal Set']
        except Exception as inst:
            return [-1, "Error: value could not be set"
                    + inst.__str__()]

    def getState(self):
        """*Returns active state of tool*.

        | *Parameters*
        |   none

        | *Returns*
        |   [1, "Tool On"]
        |   [0, "Tool Off"]
        |   [-1, "Error: Tool activation state cannot be determined + Error]
        """
        try:
            if self.dispenseStatus():
                return [1, "Tool On"]
            else:
                return [0, "Tool Off"]
        except Exception as inst:
            return [-1, "Error: Tool activation state cannot be determined"
                    + inst.__str__()]

    ##########################################################################
    ### PCP_SerialDevice METHODS
    ##########################################################################

    def checkIfSerialConnectParamsSet(self):
        """*Goes through connection parameters and sees if all are set*.

        Returns
        -------
        bool
            True
        """
        return True

    def startSerial(self):
        """*Creates and connects pySerial device*.

        Returns
        -------
        [1, "Connected Succesfully to Serial Device"]
        """
        return [1, "Connected Succesfully to Serial Device"]

    def stopSerial(self):
        """*Terminates communication*.

        Returns
        -------
        [1, "Terminated successfully"]
        """
        return [1, "Terminated successfully"]

    ################### Communication METHODS ###############################

    def handShakeSerial(self):
        """*Perform communications handshake with Tool*.

        Returns
        -------
        [1, "Handshake Successful"]
        """
        return [1, "Handshake Successful"]

    def __writeSerial__(self, text):
        """*Writes text to serial device*.

        Parameters
        ----------
        text: String
            message to send

        Returns
        -------
        [1, 'Text Sent + text']
        """
        return [1, 'Text Sent + text']

    def writeSerialCommand(self, cmdString):
        """*Writes dlcommand to serial device*.

        Parameters
        ----------
        cmdString, String
            the string to send

        Returns
        -------
        [1, 'Command Sent: ' + cmdString + 'Received: ' + rcvd]
            if exception
        """
        return [1, 'Command Sent: ' + cmdString + '\n Received: '
                + "Null tool - no receive"]

    def readTime(self):
        """*Reads in from serial device until timeout*.

        Returns
        -------
        [1, inp String of all text read in, empty string if nothing]
        """
        print('\tReceived from Serial Device: ' + 'Null device input')

##############################################################################
########################## PCP_BasicLogger METHODS ###########################
##############################################################################

    def writeLogSelf(self):
        """*Generates json string containing dict to be written to log file*.

        Returns
        -------
        logJson: String
            log in json string format
        """
        return super().writeLogSelf()

    def loadLogSelf(self, jsonString):
        """*loads json log back into dict*.

        Parameters
        ----------
        jsonString: String
            json string to be loaded back in
        """
        super().loadLogSelf(jsonString)
